//--------------------------------------------------------------------------------------------------------------------//
//                                                RTTI CLASS & MACROS                                                 //
// 03/07/01, Mythos                                                                                                   //
//--------------------------------------------------------------------------------------------------------------------//

#ifndef   _RTTI_H_
#define   _RTTI_H_
#pragma    once

//----------------------------------------------- INCLUDES -----------------------------------------------------------//

#include  "StdString.h"

//----------------------------------------------- CLASSES ------------------------------------------------------------//

//--------------------------------------------------------------------------------------------------------------------//
//                                                CRTTI                                                               //
//--------------------------------------------------------------------------------------------------------------------//

class CRTTI
  {
// public methods

  public:
                              CRTTI               (const CStdString& strClassName,const CRTTI* pBaseRTTI) :
                                                      m_strClassName(strClassName),m_pBaseRTTI(pBaseRTTI)
                                                                                {}
                             ~CRTTI               (void)                        {}

    const CStdString&         strGetClassName     (void) const                  { return m_strClassName; }
    const CRTTI*              pGetBaseRTTI        (void) const                  { return m_pBaseRTTI; }

// protected members

  protected:

    const CStdString          m_strClassName;               // name of the class containing this (static) object
    const CRTTI*              m_pBaseRTTI;                  // pointer to the (static) object of the container's base class
  };

//----------------------------------------------- MACROS -------------------------------------------------------------//

// in non-root class .h
// put this macro in the declaration of a DERIVED class - it defines a static CRTTI object for this class
// (be careful that after this macro, things are public)

#define   _DECLARE_RTTI_                                                                             \
  public:                                                                                            \
    static  const CRTTI       m_RTTI;                                                                \
    virtual const CRTTI*      pGetRTTI            (void) const                  { return &m_RTTI; }

//in root class .h
// put this macro in the declaration of a BASE class
// (be careful that after this macro, things are public)

#define   _DECLARE_ROOT_RTTI_                                                                                                 \
  public:                                                                                                                     \
    _DECLARE_RTTI_                                                                                                            \
    bool                      boIsExactClass      (const CRTTI* pRTTI)          { return(pGetRTTI() == pRTTI); }              \
    bool                      boIsKindOf          (const CRTTI* pRTTI)          { const CRTTI* pBase = pGetRTTI();            \
                                                                                  while(pBase)                                \
                                                                                    {                                         \
                                                                                    if(pBase == pRTTI) return true;           \
                                                                                    pBase = pBase->pGetBaseRTTI();            \
                                                                                    }                                         \
                                                                                  return false;                               \
                                                                                }                                             \
    void*                     pvDynamicCast       (const CRTTI* pRTTI)          { return(boIsKindOf(pRTTI) ? this : NULL); }

// in non-root class .cpp
// put this macro in the implementation file of a DERIVED class

#define _IMPLEMENT_RTTI_(ClassName,BaseClassName)                      \
  const CRTTI ClassName::m_RTTI(#ClassName,&BaseClassName::m_RTTI)

// in root class .cpp
// put this macro in the implementation file of a BASE class - it creates a static CRTTI member

#define _IMPLEMENT_ROOT_RTTI_(ClassName)                               \
  const CRTTI ClassName::m_RTTI(#ClassName,NULL)

// runtime type macros

#define _IS_EXACT_CLASS_(ClassName,pObj)                               \
  (pObj ? pObj->boIsExactClass(&ClassName::m_RTTI) : false)

#define _IS_KIND_OF_(ClassName,pObj)                                   \
    (pObj ? pObj->boIsKindOf(&ClassName::m_RTTI) : false)

#define _STATIC_CAST_(ClassName,pObj)                                  \
  ((ClassName*)pObj)

#define _DYNAMIC_CAST_(ClassName,pObj)                                 \
  (pObj ? _STATIC_CAST_(ClassName,pObj->pvDynamicCast(&ClassName::m_RTTI)) : NULL)

//--------------------------------------------------------------------------------------------------------------------//

#endif // _RTTI_H_
