//--------------------------------------------------------------------------------------------------------------------//
//                                                SMART POINTER TEMPLATE                                              //
// 18/06/00, Mythos                                                                                                   //
//--------------------------------------------------------------------------------------------------------------------//

#ifndef   _SMARTPTR_INL_
#define   _SMARTPTR_INL_

//--------------------------------------------------------------------------------------------------------------------//
//                                                CONSTRUCTORS / DESTRUCTOR                                           //
//--------------------------------------------------------------------------------------------------------------------//

template <class CObj> inline CSmartPtr<CObj>::CSmartPtr(CObj* pObj)
  {
  m_pDumbPtr = pObj;
  if(m_pDumbPtr) m_pDumbPtr->vAddRef();
  }

//

template <class CObj> inline CSmartPtr<CObj>::CSmartPtr(const CSmartPtr& Smart)
  {
  m_pDumbPtr = Smart.m_pDumbPtr;
  if(m_pDumbPtr) m_pDumbPtr->vAddRef();
  }

//

template <class CObj> inline CSmartPtr<CObj>::~CSmartPtr()
  {
  if(m_pDumbPtr) m_pDumbPtr->vReleaseRef();
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                OPERATORS                                                           //
//--------------------------------------------------------------------------------------------------------------------//

template <class CObj> inline CSmartPtr<CObj>& CSmartPtr<CObj>::operator = (const CSmartPtr& Smart)
  {
  if(m_pDumbPtr != Smart.m_pDumbPtr)                        // check assignment to self
    {
    if(m_pDumbPtr) m_pDumbPtr->vReleaseRef();

    m_pDumbPtr = Smart.m_pDumbPtr;
    if(m_pDumbPtr) m_pDumbPtr->vAddRef();
    }

  return *this;
  }

//

template <class CObj> inline CSmartPtr<CObj>& CSmartPtr<CObj>::operator = (CObj* pObj)
  {
  if(m_pDumbPtr != pObj)                                    // check assignment to self
    {
    if(m_pDumbPtr) m_pDumbPtr->vReleaseRef();

    m_pDumbPtr = pObj;
    if(m_pDumbPtr) m_pDumbPtr->vAddRef();
    }

  return *this;
  }

//

template <class CObj> inline CObj* CSmartPtr<CObj>::operator -> () const
  {
  return m_pDumbPtr;
  }

//

template <class CObj> inline CObj& CSmartPtr<CObj>::operator * () const
  {
  return *m_pDumbPtr;
  }

//

template <class CObj> inline CSmartPtr<CObj>::operator CObj* () const
  {
  return m_pDumbPtr;
  }

//

template <class CObj> inline bool CSmartPtr<CObj>::operator == (CObj* pObj) const
  {
  return(m_pDumbPtr == pObj);
  }

//

template <class CObj> inline bool CSmartPtr<CObj>::operator != (CObj* pObj) const
  {
  return(m_pDumbPtr != pObj);
  }

//

template <class CObj> inline bool CSmartPtr<CObj>::operator == (const CSmartPtr& Smart) const
  {
  return(m_pDumbPtr == Smart.m_pDumbPtr);
  }

//

template <class CObj> inline bool CSmartPtr<CObj>::operator != (const CSmartPtr& Smart) const
  {
  return(m_pDumbPtr != Smart.m_pDumbPtr);
  }

// this can sometimes be useful to cast the dumb pointer :

template <class CObj> inline CObj* CSmartPtr<CObj>::pGetDumb() const
  {
  return m_pDumbPtr;
  }

//--------------------------------------------------------------------------------------------------------------------//

#endif // _SMARTPTR_INL_
