//--------------------------------------------------------------------------------------------------------------------//
//                                                VECTORS & MATRICES CLASSES                                          //
// 17/09/97 - 09/06/00, Mythos                                                                                        //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- INCLUDES -----------------------------------------------------------//

#include  "stdafx.h"
#include  "VectMat.h"

//--------------------------------------------------------------------------------------------------------------------//
//                                                3*3 MATRIX                                                          //
//--------------------------------------------------------------------------------------------------------------------//

CMat3x3 CMat3x3::m3Inverse() const
  {                                                                   // Gauss-Jordan elimination with partial pivoting
  CMat3x3 m3A(*this);                                                 // as A evolves from the original  matrix to identity
  CMat3x3 m3B(m3Identity3D());                                        // ...B evolves from identity to inverse(A)
  BYTE    bI,bJ,bI1;

  // loop over cols of A from left to right, eliminating above & below diagonals

  for(bJ = 0; bJ < 3; bJ++)                                           // find largest pivot in column j among rows j..2
    { 
    bI1 = bJ;                                                         // row with largest pivot candidate
    for(bI = BYTE(bJ+1); bI < 3; bI++) if(fabs(m3A.m_v3V[bI].m_fV[bJ]) > fabs(m3A.m_v3V[bI1].m_fV[bJ])) bI1 = bI;

    // swap rows i1 & j in A & B to put pivot on diagonal
  
    vSwap(m3A.m_v3V[bI1], m3A.m_v3V[bJ]);
    vSwap(m3B.m_v3V[bI1], m3B.m_v3V[bJ]);

    // scale row j to have a unit diagonal
    
    float fK = m3A.m_v3V[bJ].m_fV[bJ];
    _DDIV0_(fK)
    float fKinv = 1.0f/fK;
    m3B.m_v3V[bJ] *= fKinv;
    m3A.m_v3V[bJ] *= fKinv;

    // eliminate off-diagonal elements in col j of A, doing identical ops to B
    
    for(bI = 0; bI < 3; bI++) if(bI != bJ)
      { 
      m3B.m_v3V[bI] -= m3A.m_v3V[bI].m_fV[bJ]*m3B.m_v3V[bJ];
      m3A.m_v3V[bI] -= m3A.m_v3V[bI].m_fV[bJ]*m3A.m_v3V[bJ]; 
      }
    }

  return m3B;
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                4*4 MATRIX                                                          //
//--------------------------------------------------------------------------------------------------------------------//

CMat4x4 CMat4x4::m4Inverse() const
  {                                                                   // Gauss-Jordan elimination with partial pivoting
  CMat4x4 m4A(*this);                                                 // as A evolves from the original  matrix to identity
  CMat4x4 m4B(m4Identity4D());                                        // ...B evolves from identity to inverse(A)
  BYTE    bI,bJ,bI1;

  // loop over cols of A from left to right, eliminating above & below diagonals

  for(bJ = 0; bJ < 4; bJ++)                                           // find largest pivot in column j among rows j..3
    { 
    bI1 = bJ;                                                         // row with largest pivot candidate
    for(bI = BYTE(bJ+1); bI < 4; bI++) if(fabs(m4A.m_v4V[bI].m_fV[bJ]) > fabs(m4A.m_v4V[bI1].m_fV[bJ])) bI1 = bI;

    // swap rows i1 & j in A & B to put pivot on diagonal
  
    vSwap(m4A.m_v4V[bI1], m4A.m_v4V[bJ]);
    vSwap(m4B.m_v4V[bI1], m4B.m_v4V[bJ]);

    // scale row j to have a unit diagonal
    
    float fK = m4A.m_v4V[bJ].m_fV[bJ];
    _DDIV0_(fK);
    float fKinv = 1.0f/fK;
    m4B.m_v4V[bJ] *= fKinv;
    m4A.m_v4V[bJ] *= fKinv;

    // eliminate off-diagonal elements in col j of A, doing identical ops to B
    
    for(bI = 0; bI < 4; bI++) if(bI != bJ)
      { 
      m4B.m_v4V[bI] -= m4A.m_v4V[bI].m_fV[bJ]*m4B.m_v4V[bJ];
      m4A.m_v4V[bI] -= m4A.m_v4V[bI].m_fV[bJ]*m4A.m_v4V[bJ]; 
      }
    }

    return m4B;
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                TRANSFORMATIONS                                                     //
//--------------------------------------------------------------------------------------------------------------------//

//--------------------------------------------------------------------------------------------------------------------//
//                                                basic transformations                                               //
// translation, rotation, scale, perspective                                                                          //
//--------------------------------------------------------------------------------------------------------------------//

CMat3x3 m3Translation2D(const CVect2D& v2V)                           // 2D translation (3*3 matrix)
  { return CMat3x3(CVect3D(1.f, 0.f, v2V[_X_]),
                   CVect3D(0.f, 1.f, v2V[_Y_]),
                   CVect3D(0.f, 0.f, 1.f)); }

CMat4x4 m4Translation3D(const CVect3D& m_v3V)                         // 3D translation (4*4 matrix)
  { return CMat4x4(CVect4D(1.f, 0.f, 0.f, m_v3V[_X_]),
                   CVect4D(0.f, 1.f, 0.f, m_v3V[_Y_]),
                   CVect4D(0.f, 0.f, 1.f, m_v3V[_Z_]),
                   CVect4D(0.f, 0.f, 0.f, 1.f)); }

CMat3x3 m3Rotation2D(const CVect2D& v2Center,const float fAngleRad)
  {                                                                   // 2D rotation around a point (3*3 matrix)
  float fCos = (float)cos(fAngleRad);
  float fSin = (float)sin(fAngleRad);
  float fT   = (float)1.f-fCos;

  return CMat3x3(CVect3D(fCos, -fSin, v2Center[_X_]*fT+v2Center[_Y_]*fSin),
                 CVect3D(fSin,  fCos, v2Center[_Y_]*fT-v2Center[_X_]*fSin),
                 CVect3D(0.f,   0.f,  1.f)); 
  }

CMat4x4 m4Rotation3D(const CVect3D& v3Axis,const float fAngleRad)       
  {                                                                   // 3D rotation around an axis (4*4 matrix)
  float fCos = (float)cos(fAngleRad);
  float fSin = (float)sin(fAngleRad);
  float fT   = (float)1.f-fCos;
  float fSX  = v3Axis[_X_]*fSin, fSY = v3Axis[_Y_]*fSin, fSZ = v3Axis[_Z_]*fSin;
  float fTX  = v3Axis[_X_]*fT,   fTY = v3Axis[_Y_]*fT,   fTZ = v3Axis[_Z_]*fT;

  const_cast<CVect3D&>(v3Axis).boNormalize();

  return CMat4x4(CVect4D(fTX*v3Axis[_X_]+fCos, fTX*v3Axis[_Y_]-fSZ,  fTX*v3Axis[_Z_]+fSY,  0.f),
                 CVect4D(fTX*v3Axis[_Y_]+fSZ,  fTY*v3Axis[_Y_]+fCos, fTY*v3Axis[_Z_]-fSX,  0.f),
                 CVect4D(fTX*v3Axis[_Z_]-fSY,  fTY*v3Axis[_Z_]+fSX,  fTZ*v3Axis[_Z_]+fCos, 0.f),
                 CVect4D(       0.f,                  0.f,                  0.f,           1.f)); 
  }

CMat3x3 m3Scale2D(const CVect2D& v2Scale)                             // 2D scale (3*3 matrix)
  { return CMat3x3(CVect3D(v2Scale[_X_], 0.f, 0.f),
                   CVect3D(0.f, v2Scale[_Y_], 0.f),
                   CVect3D(0.f, 0.f, 1.f)); }

CMat4x4 m4Scale3D(const CVect3D& v3Scale)                             // 3D scale (4*4 matrix)
  { return CMat4x4(CVect4D(v3Scale[_X_], 0.f, 0.f, 0.f),
                   CVect4D(0.f, v3Scale[_Y_], 0.f, 0.f),
                   CVect4D(0.f, 0.f, v3Scale[_Z_], 0.f),
                   CVect4D(0.f, 0.f, 0.f, 1.f)); }

CMat4x4 m4Perspective3D(const float fD)                               // 3D perspective (4*4 matrix)
  {
  _DDIV0_(fD);
  return CMat4x4(CVect4D(1.f, 0.f, 0.f, 0.f),
                 CVect4D(0.f, 1.f, 0.f, 0.f),
                 CVect4D(0.f, 0.f, 1.f, 0.f),
                 CVect4D(0.f, 0.f, (float)1.f/fD, 0.f)); 
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                CALCULATIONS                                                        //
//--------------------------------------------------------------------------------------------------------------------//

//----------------------------------------------- vMultHomoMatrix ----------------------------------------------------//
// multiply two homogeneous matrices (last line = 0;0;0;1)
// 25/12/97, Mythos
// in : result,M1,M2
// out: result=M1*M2
//--------------------------------------------------------------------------------------------------------------------//

void vMultHomoMatrix(CMat4x4* pm4M,const CMat4x4& m4M1,const CMat4x4& m4M2)
  {
  (*pm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[0][1]*m4M2[1][0]
                + m4M1[0][2]*m4M2[2][0];
  (*pm4M)[0][1] = m4M1[0][0]*m4M2[0][1] + m4M1[0][1]*m4M2[1][1]
                + m4M1[0][2]*m4M2[2][1];
  (*pm4M)[0][2] = m4M1[0][0]*m4M2[0][2] + m4M1[0][1]*m4M2[1][2]
                + m4M1[0][2]*m4M2[2][2];
  (*pm4M)[0][3] = m4M1[0][0]*m4M2[0][3] + m4M1[0][1]*m4M2[1][3]
                + m4M1[0][2]*m4M2[2][3] + m4M1[0][3];

  (*pm4M)[1][0] = m4M1[1][0]*m4M2[0][0] + m4M1[1][1]*m4M2[1][0]
                + m4M1[1][2]*m4M2[2][0];
  (*pm4M)[1][1] = m4M1[1][0]*m4M2[0][1] + m4M1[1][1]*m4M2[1][1]
                + m4M1[1][2]*m4M2[2][1];
  (*pm4M)[1][2] = m4M1[1][0]*m4M2[0][2] + m4M1[1][1]*m4M2[1][2]
                + m4M1[1][2]*m4M2[2][2];
  (*pm4M)[1][3] = m4M1[1][0]*m4M2[0][3] + m4M1[1][1]*m4M2[1][3]
                + m4M1[1][2]*m4M2[2][3] + m4M1[1][3];

  (*pm4M)[2][0] = m4M1[2][0]*m4M2[0][0] + m4M1[2][1]*m4M2[1][0]
                + m4M1[2][2]*m4M2[2][0];
  (*pm4M)[2][1] = m4M1[2][0]*m4M2[0][1] + m4M1[2][1]*m4M2[1][1]
                + m4M1[2][2]*m4M2[2][1];
  (*pm4M)[2][2] = m4M1[2][0]*m4M2[0][2] + m4M1[2][1]*m4M2[1][2]
                + m4M1[2][2]*m4M2[2][2];
  (*pm4M)[2][3] = m4M1[2][0]*m4M2[0][3] + m4M1[2][1]*m4M2[1][3]
                + m4M1[2][2]*m4M2[2][3] + m4M1[2][3];

  (*pm4M)[3][0] = (*pm4M)[3][1] = (*pm4M)[3][2] = 0.f;
  (*pm4M)[3][3] = 1.f;
  }

//----------------------------------------------- vTransposeHomoMatrix -----------------------------------------------//
// transpose a 4*4 transformation (transpose rotation,change translation sign)
// 25/12/97, Mythos
// in : result,M
// out: result=Mt
// rem: doesn't work if there are scale factors in the matrix (the scale is not reversed)
//--------------------------------------------------------------------------------------------------------------------//

void vTransposeHomoMatrix(CMat4x4* pm4Mt,const CMat4x4& m4M)
  {
  (*pm4Mt)[0][0] = m4M[0][0];                                         // transpose rotation
  (*pm4Mt)[0][1] = m4M[1][0];
  (*pm4Mt)[0][2] = m4M[2][0];
  (*pm4Mt)[1][0] = m4M[0][1];
  (*pm4Mt)[1][1] = m4M[1][1];
  (*pm4Mt)[1][2] = m4M[2][1];
  (*pm4Mt)[2][0] = m4M[0][2];
  (*pm4Mt)[2][1] = m4M[1][2];
  (*pm4Mt)[2][2] = m4M[2][2];

  (*pm4Mt)[0][3] =-m4M[0][3];                                         // reverse translation
  (*pm4Mt)[1][3] =-m4M[1][3];
  (*pm4Mt)[2][3] =-m4M[2][3];

  (*pm4Mt)[3][0] = (*pm4Mt)[3][1] = (*pm4Mt)[3][2] = 0.f;             // last line
  (*pm4Mt)[3][3] = 1.f;
  }

//----------------------------------------------- vRotate3DVector4x4 -------------------------------------------------//
// use the rotation part of a 4*4 matrix to rotate a 3*3 vector
// 25/12/97, Mythos
// in : result,Matrix, Vector
// out: result=rotated Vector
//--------------------------------------------------------------------------------------------------------------------//

void vRotate3DVector4x4(CVect3D* pv3Vp,const CMat4x4& m4M,const CVect3D& v3V)
  {
  (*pv3Vp)[_X_] = m4M[0][0]*v3V[_X_] + m4M[0][1]*v3V[_Y_] + m4M[0][2]*v3V[_Z_];
  (*pv3Vp)[_Y_] = m4M[1][0]*v3V[_X_] + m4M[1][1]*v3V[_Y_] + m4M[1][2]*v3V[_Z_];
  (*pv3Vp)[_Z_] = m4M[2][0]*v3V[_X_] + m4M[2][1]*v3V[_Y_] + m4M[2][2]*v3V[_Z_];
  }

//----------------------------------------------- vCalcRotMatrix -----------------------------------------------------//
// calculate 4x4 rotation matrix from 3 angles (Bryant)
// 25/12/97, Mythos
// in : result,angles
// out:
//--------------------------------------------------------------------------------------------------------------------//

void vCalcRotMatrix(CMat4x4* pm4M,const CVect3D& v3Angles)
  {
  float c0,s0,c1,s1,c2,s2;
  c0 = (float)cos(v3Angles[_X_]); s0 = (float)sin(v3Angles[_X_]);
  c1 = (float)cos(v3Angles[_Y_]); s1 = (float)sin(v3Angles[_Y_]);
  c2 = (float)cos(v3Angles[_Z_]); s2 = (float)sin(v3Angles[_Z_]);
  (*pm4M)[2][0] = -s1;

  float c0c1 = c0*c1;
  float s0c1 = s0*c1;
  (*pm4M)[2][2] = c0c1;
  (*pm4M)[2][1] = s0c1;

  float c1c2 = c1*c2;
  float c1s2 = c1*s2;
  (*pm4M)[0][0] = c1c2;
  (*pm4M)[1][0] = c1s2;

  float c0s2 = c0*s2;
  float c0c2 = c0*c2;
  float s0s2 = s0*s2;
  float s0c2 = s0*c2;
  (*pm4M)[0][1] =-c0s2+s0c2*s1;
  (*pm4M)[1][1] = c0c2+s0s2*s1;
  (*pm4M)[0][2] = s0s2+c0c2*s1;
  (*pm4M)[1][2] =-s0c2+c0s2*s1;
  }

//----------------------------------------------- vCalcRotMatrix -----------------------------------------------------//
// calculate 3x3 rotation matrix from 3 angles (Bryant)
// 01/10/99, Mythos
// in : result,angles
// out:
//--------------------------------------------------------------------------------------------------------------------//

void vCalcRotMatrix(CMat3x3* pm3M,const CVect3D& v3Angles)
  {
  float c0,s0,c1,s1,c2,s2;
  c0 = (float)cos(v3Angles[_X_]); s0 = (float)sin(v3Angles[_X_]);
  c1 = (float)cos(v3Angles[_Y_]); s1 = (float)sin(v3Angles[_Y_]);
  c2 = (float)cos(v3Angles[_Z_]); s2 = (float)sin(v3Angles[_Z_]);
  (*pm3M)[2][0] = -s1;

  float c0c1 = c0*c1;
  float s0c1 = s0*c1;
  (*pm3M)[2][2] = c0c1;
  (*pm3M)[2][1] = s0c1;

  float c1c2 = c1*c2;
  float c1s2 = c1*s2;
  (*pm3M)[0][0] = c1c2;
  (*pm3M)[1][0] = c1s2;

  float c0s2 = c0*s2;
  float c0c2 = c0*c2;
  float s0s2 = s0*s2;
  float s0c2 = s0*c2;
  (*pm3M)[0][1] =-c0s2+s0c2*s1;
  (*pm3M)[1][1] = c0c2+s0s2*s1;
  (*pm3M)[0][2] = s0s2+c0c2*s1;
  (*pm3M)[1][2] =-s0c2+c0s2*s1;
}

//--------------------------------------------------------------------------------------------------------------------//
//                                                2D                                                                  //
//--------------------------------------------------------------------------------------------------------------------//

void vAdd(CVect2D* pv2V, const CVect2D& v2V1)                         // V+=V1
  {
  (*pv2V)[0] += v2V1[0];
  (*pv2V)[1] += v2V1[1];
  }

void vSub(CVect2D* pv2V, const CVect2D& v2V1)                         // V-=V1
  {
  (*pv2V)[0] -= v2V1[0];
  (*pv2V)[1] -= v2V1[1];
  }

void vMul(CVect2D* pv2V, const float fK)                              // V*=k
  {
  (*pv2V)[0] *= fK;
  (*pv2V)[1] *= fK;
  }

void vMul(CVect2D* pv2V, const CVect2D& v2V1)                         // V*=V1 (term by term)
  {
  (*pv2V)[0] *= v2V1[0];
  (*pv2V)[1] *= v2V1[1];
  }

void vDiv(CVect2D* pv2V, const float fK)                              // V/=k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;
  (*pv2V)[0] *= fKinv;
  (*pv2V)[1] *= fKinv;
  }

void vDiv(CVect2D* pv2V, const CVect2D& v2V1)                         // V/=V1 (term by term)
  {
  _DDIV0_(v2V1[0]);
  _DDIV0_(v2V1[1]);
  (*pv2V)[0] /= v2V1[0];
  (*pv2V)[1] /= v2V1[1];
  }

void vNeg(CVect2D* pv2V, const CVect2D& v2V1)                         // V=-V1
  {
  (*pv2V)[0] = -v2V1[0];
  (*pv2V)[1] = -v2V1[1];
  }

void vMul(CVect2D* pv2V, const CVect2D& v2V1,const float fK)
  {                                                                   // V=V1*k
  (*pv2V)[0] = v2V1[0]*fK;
  (*pv2V)[1] = v2V1[1]*fK;
  }

void vDiv(CVect2D* pv2V, const CVect2D& v2V1,const float fK)
  {                                                                   // V=V1/k
  _DDIV0_(fK);
  float fKinv = 1.f/fK;
  (*pv2V)[0] = v2V1[0]*fKinv;
  (*pv2V)[1] = v2V1[1]*fKinv;
  }

void vSub(CVect2D* pv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1-V2
  (*pv2V)[0] = v2V1[0] - v2V2[0];
  (*pv2V)[1] = v2V1[1] - v2V2[1];
  }

void vAdd(CVect2D* pv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1+V2
  (*pv2V)[0] = v2V1[0] + v2V2[0];
  (*pv2V)[1] = v2V1[1] + v2V2[1];
  }

float fDotProd  (const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // f=V1*V2
  return(v2V1[0]*v2V2[0] + v2V1[1]*v2V2[1]);
  }

float fCrossProd(const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // f=V1^V2
  return(v2V1[0]*v2V2[1] - v2V1[1]*v2V2[0]);
  }

void vMul(CVect2D* pv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1*V2 (term by term)
  (*pv2V)[0] = v2V1[0] * v2V2[0];
  (*pv2V)[1] = v2V1[1] * v2V2[1];
  }

void vDiv(CVect2D* pv2V, const CVect2D& v2V1,const CVect2D& v2V2)
  {                                                                   // V=V1/V2 (term by term)
  _DDIV0_(v2V2[0]);
  _DDIV0_(v2V2[1]);
  (*pv2V)[0] = v2V1[0] / v2V2[0];
  (*pv2V)[1] = v2V1[1] / v2V2[1];
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                3D                                                                  //
//--------------------------------------------------------------------------------------------------------------------//

void vAdd(CVect3D* pv3V, const CVect3D& v3V1)                         // V+=V1
  {
  (*pv3V)[0] += v3V1[0];
  (*pv3V)[1] += v3V1[1];
  (*pv3V)[2] += v3V1[2];
  }
          
void vSub(CVect3D* pv3V, const CVect3D& v3V1)                         // V-=V1
  {
  (*pv3V)[0] -= v3V1[0];
  (*pv3V)[1] -= v3V1[1];
  (*pv3V)[2] -= v3V1[2];
  }          

void vMul(CVect3D* pv3V, const float fK)                              // V*=k
  {
  (*pv3V)[0] *= fK;
  (*pv3V)[1] *= fK;
  (*pv3V)[2] *= fK;          
  }

void vMul(CVect3D* pv3V, const CVect3D& v3V1)                         // V*=V1 (term by term)
  {
  (*pv3V)[0] *= v3V1[0];
  (*pv3V)[1] *= v3V1[1];          
  (*pv3V)[2] *= v3V1[2];
  }

void vDiv(CVect3D* pv3V, const float fK)                              // V/=k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;
  (*pv3V)[0] *= fKinv;          
  (*pv3V)[1] *= fKinv;
  (*pv3V)[2] *= fKinv;
  }

void vDiv(CVect3D* pv3V, const CVect3D& v3V1)                         // V/=V1 (term by term)
  {
  _DDIV0_(v3V1[0]);
  _DDIV0_(v3V1[1]);          
  _DDIV0_(v3V1[2]);
  (*pv3V)[0] /= v3V1[0];
  (*pv3V)[1] /= v3V1[1];
  (*pv3V)[2] /= v3V1[2];
  }

void vNeg(CVect3D* pv3V, const CVect3D& v3V1)                         // V=-V1
  {
  (*pv3V)[0] = -v3V1[0];
  (*pv3V)[1] = -v3V1[1];
  (*pv3V)[2] = -v3V1[2];
  }

void vMul(CVect3D* pv3V, const CVect3D& v3V1,const float fK)
  {                                                                   // V=V1*k
  (*pv3V)[0] = v3V1[0]*fK;
  (*pv3V)[1] = v3V1[1]*fK;
  (*pv3V)[2] = v3V1[2]*fK;
  }

void vDiv(CVect3D* pv3V, const CVect3D& v3V1,const float fK)
  {                                                                   // V=V1/k
  _DDIV0_(fK);
  float fKinv = 1.f/fK;
  (*pv3V)[0] = v3V1[0]*fKinv;
  (*pv3V)[1] = v3V1[1]*fKinv;
  (*pv3V)[2] = v3V1[2]*fKinv;
  }

void vSub(CVect3D* pv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1-V2
  (*pv3V)[0] = v3V1[0] - v3V2[0];
  (*pv3V)[1] = v3V1[1] - v3V2[1];
  (*pv3V)[2] = v3V1[2] - v3V2[2];
  }

void vAdd(CVect3D* pv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1+V2
  (*pv3V)[0] = v3V1[0] + v3V2[0];
  (*pv3V)[1] = v3V1[1] + v3V2[1];
  (*pv3V)[2] = v3V1[2] + v3V2[2];
  }

float fDotProd(const CVect3D& v3V1,const CVect3D& v3V2)               // f=V1*V2
  {
  return(v3V1[0]*v3V2[0] + v3V1[1]*v3V2[1] + v3V1[2]*v3V2[2]);
  }

float fDotProd(const CVect4D& v4V1,const CVect3D& v3V2)               // f=V1*V2
  {
  return(v4V1[0]*v3V2[0] + v4V1[1]*v3V2[1] + v4V1[2]*v3V2[2] + v4V1[3]);
  }

void vCrossProd(CVect3D* pv3V,const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1^V2
  (*pv3V)[0] = v3V1[_Y_]*v3V2[_Z_] - v3V1[_Z_]*v3V2[_Y_];
  (*pv3V)[1] = v3V1[_Z_]*v3V2[_X_] - v3V1[_X_]*v3V2[_Z_];
  (*pv3V)[2] = v3V1[_X_]*v3V2[_Y_] - v3V1[_Y_]*v3V2[_X_];
  }

void vMul(CVect3D* pv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1*V2 (term by term)
  (*pv3V)[0] = v3V1[0] * v3V2[0];
  (*pv3V)[1] = v3V1[1] * v3V2[1];
  (*pv3V)[2] = v3V1[2] * v3V2[2];
  }

void vDiv(CVect3D* pv3V, const CVect3D& v3V1,const CVect3D& v3V2)
  {                                                                   // V=V1/V2 (term by term)
  _DDIV0_(v3V2[0]);
  _DDIV0_(v3V2[1]);
  _DDIV0_(v3V2[2]);
  (*pv3V)[0] = v3V1[0] / v3V2[0];
  (*pv3V)[1] = v3V1[1] / v3V2[1];
  (*pv3V)[2] = v3V1[2] / v3V2[2];
  }

void vMul(CVect3D* pv3V, const CMat3x3& m3M ,const CVect3D& v3V1)     // V=M*V1
  {
  (*pv3V)[0] = m3M[0][0]*v3V1[0] + m3M[0][1]*v3V1[1] + m3M[0][2]*v3V1[2];
  (*pv3V)[1] = m3M[1][0]*v3V1[0] + m3M[1][1]*v3V1[1] + m3M[1][2]*v3V1[2];
  (*pv3V)[2] = m3M[2][0]*v3V1[0] + m3M[2][1]*v3V1[1] + m3M[2][2]*v3V1[2];
  }

void vMul(CVect3D* pv3V, const CVect3D& v3V1,const CMat3x3& m3M)      // V=V1*M
  {
  (*pv3V)[0] = m3M[0][0]*v3V1[0] + m3M[1][0]*v3V1[1] + m3M[2][0]*v3V1[2];
  (*pv3V)[1] = m3M[0][1]*v3V1[0] + m3M[1][1]*v3V1[1] + m3M[2][1]*v3V1[2];
  (*pv3V)[2] = m3M[0][2]*v3V1[0] + m3M[1][2]*v3V1[1] + m3M[2][2]*v3V1[2];
  }

void vMul(CVect4D* pv4V, const CMat4x4& m4M ,const CVect3D& v3V1)     // V=M*V1
  {
  (*pv4V)[0] = m4M[0][0]*v3V1[0] + m4M[0][1]*v3V1[1] + m4M[0][2]*v3V1[2] + m4M[0][3];
  (*pv4V)[1] = m4M[1][0]*v3V1[0] + m4M[1][1]*v3V1[1] + m4M[1][2]*v3V1[2] + m4M[1][3];
  (*pv4V)[2] = m4M[2][0]*v3V1[0] + m4M[2][1]*v3V1[1] + m4M[2][2]*v3V1[2] + m4M[2][3];
  (*pv4V)[3] = m4M[3][0]*v3V1[0] + m4M[3][1]*v3V1[1] + m4M[3][2]*v3V1[2] + m4M[3][3];
  }

void vMul(CVect4D* pv4V, const CVect3D& v3V1,const CMat4x4& m4M)      // V=V1*M
  {
  (*pv4V)[0] = m4M[0][0]*v3V1[0] + m4M[1][0]*v3V1[1] + m4M[2][0]*v3V1[2] + m4M[3][0];
  (*pv4V)[1] = m4M[0][1]*v3V1[0] + m4M[1][1]*v3V1[1] + m4M[2][1]*v3V1[2] + m4M[3][1];
  (*pv4V)[2] = m4M[0][2]*v3V1[0] + m4M[1][2]*v3V1[1] + m4M[2][2]*v3V1[2] + m4M[3][2];
  (*pv4V)[3] = m4M[0][3]*v3V1[0] + m4M[1][3]*v3V1[1] + m4M[2][3]*v3V1[2] + m4M[3][3];
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                4D                                                                  //
//--------------------------------------------------------------------------------------------------------------------//

void vAdd(CVect4D* pv4V, const CVect4D& v4V1)                         // V+=V1
  {
  (*pv4V)[0] += v4V1[0];
  (*pv4V)[1] += v4V1[1];
  (*pv4V)[2] += v4V1[2];
  (*pv4V)[3] += v4V1[3];
  }

void vSub(CVect4D* pv4V, const CVect4D& v4V1)                         // V-=V1
  {
  (*pv4V)[0] -= v4V1[0];
  (*pv4V)[1] -= v4V1[1];
  (*pv4V)[2] -= v4V1[2];
  (*pv4V)[3] -= v4V1[3];
  }

void vMul(CVect4D* pv4V, const float fK)                              // V*=k
  {
  (*pv4V)[0] *= fK;
  (*pv4V)[1] *= fK;
  (*pv4V)[2] *= fK;
  (*pv4V)[3] *= fK;
  }

void vMul(CVect4D* pv4V, const CVect4D& v4V1)                         // V*=V1 (term by term)
  {
  (*pv4V)[0] *= v4V1[0];
  (*pv4V)[1] *= v4V1[1];
  (*pv4V)[2] *= v4V1[2];
  (*pv4V)[3] *= v4V1[3];
  }

void vDiv(CVect4D* pv4V, const float fK)                              // V/=k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;
  (*pv4V)[0] *= fKinv;
  (*pv4V)[1] *= fKinv;
  (*pv4V)[2] *= fKinv;
  (*pv4V)[3] *= fKinv;
  }

void vDiv(CVect4D* pv4V, const CVect4D& v4V1)                         // V/=V1 (term by term)
  {
  _DDIV0_(v4V1[0]);
  _DDIV0_(v4V1[1]);
  _DDIV0_(v4V1[2]);
  _DDIV0_(v4V1[3]);
  (*pv4V)[0] /= v4V1[0];
  (*pv4V)[1] /= v4V1[1];
  (*pv4V)[2] /= v4V1[2];
  (*pv4V)[3] /= v4V1[3];
  }

void vNeg(CVect4D* pv4V, const CVect4D& v4V1)                         // V=-V1
  {
  (*pv4V)[0] = -v4V1[0];
  (*pv4V)[1] = -v4V1[1];
  (*pv4V)[2] = -v4V1[2];
  (*pv4V)[3] = -v4V1[3];
  }

void vMul(CVect4D* pv4V, const CVect4D& v4V1,float fK)                // V=V1*k
  {
  (*pv4V)[0] = v4V1[0]*fK;
  (*pv4V)[1] = v4V1[1]*fK;
  (*pv4V)[2] = v4V1[2]*fK;
  (*pv4V)[3] = v4V1[3]*fK;
  }

void vDiv(CVect4D* pv4V, const CVect4D& v4V1,const float fK)          // V=V1/k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;
  (*pv4V)[0] = v4V1[0]*fKinv;
  (*pv4V)[1] = v4V1[1]*fKinv;
  (*pv4V)[2] = v4V1[2]*fKinv;
  (*pv4V)[3] = v4V1[3]*fKinv;
  }

void vSub(CVect4D* pv4V, const CVect4D& v4V1,const CVect4D& v4V2)     // V=V1-V2
  {
  (*pv4V)[0] = v4V1[0] - v4V2[0];
  (*pv4V)[1] = v4V1[1] - v4V2[1];
  (*pv4V)[2] = v4V1[2] - v4V2[2];
  (*pv4V)[3] = v4V1[3] - v4V2[3];
  }

void vAdd(CVect4D* pv4V, const CVect4D& v4V1,const CVect4D& v4V2)     // V=V1+V2
  {
  (*pv4V)[0] = v4V1[0] + v4V2[0];
  (*pv4V)[1] = v4V1[1] + v4V2[1];
  (*pv4V)[2] = v4V1[2] + v4V2[2];
  (*pv4V)[3] = v4V1[3] + v4V2[3];
  }

float fDotProd(const CVect4D& v4V1,const CVect4D& v4V2)               // f=V1*V2
  {
  return(v4V1[0]*v4V2[0] + v4V1[1]*v4V2[1] + v4V1[2]*v4V2[2] + v4V1[3]*v4V2[3]);
  }

void vMul(CVect4D* pv4V, const CVect4D& v4V1,const CVect4D& v4V2)     // V=V1*V2 (term by term)
  {
  (*pv4V)[0] = v4V1[0] * v4V2[0];
  (*pv4V)[1] = v4V1[1] * v4V2[1];
  (*pv4V)[2] = v4V1[2] * v4V2[2];
  (*pv4V)[3] = v4V1[3] * v4V2[3];
  }

void vDiv(CVect4D* pv4V, const CVect4D& v4V1,const CVect4D& v4V2)     // V=V1/V2 (term by term)
  {
  _DDIV0_(v4V2[0]);
  _DDIV0_(v4V2[1]);
  _DDIV0_(v4V2[2]);
  _DDIV0_(v4V2[3]);
  (*pv4V)[0] = v4V1[0] / v4V2[0];
  (*pv4V)[1] = v4V1[1] / v4V2[1];
  (*pv4V)[2] = v4V1[2] / v4V2[2];
  (*pv4V)[3] = v4V1[3] / v4V2[3];
  }

void vMul(CVect4D* pv4V, const CMat4x4& m4M ,const CVect4D& v4V1)     // V=M*V1
  {
  (*pv4V)[0] = m4M[0][0]*v4V1[0] + m4M[0][1]*v4V1[1] + m4M[0][2]*v4V1[2] + m4M[0][3]*v4V1[3];
  (*pv4V)[1] = m4M[1][0]*v4V1[0] + m4M[1][1]*v4V1[1] + m4M[1][2]*v4V1[2] + m4M[1][3]*v4V1[3];
  (*pv4V)[2] = m4M[2][0]*v4V1[0] + m4M[2][1]*v4V1[1] + m4M[2][2]*v4V1[2] + m4M[2][3]*v4V1[3];
  (*pv4V)[3] = m4M[3][0]*v4V1[0] + m4M[3][1]*v4V1[1] + m4M[3][2]*v4V1[2] + m4M[3][3]*v4V1[3];
  }

void vMul(CVect4D* pv4V, const CVect4D& v4V1,const CMat4x4& m4M)      // V=V1*M
  {
  (*pv4V)[0] = m4M[0][0]*v4V1[0] + m4M[1][0]*v4V1[1] + m4M[2][0]*v4V1[2] + m4M[3][0]*v4V1[3];
  (*pv4V)[1] = m4M[0][1]*v4V1[0] + m4M[1][1]*v4V1[1] + m4M[2][1]*v4V1[2] + m4M[3][1]*v4V1[3];
  (*pv4V)[2] = m4M[0][2]*v4V1[0] + m4M[1][2]*v4V1[1] + m4M[2][2]*v4V1[2] + m4M[3][2]*v4V1[3];
  (*pv4V)[3] = m4M[0][3]*v4V1[0] + m4M[1][3]*v4V1[1] + m4M[2][3]*v4V1[2] + m4M[3][3]*v4V1[3];
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                3x3                                                                 //
//--------------------------------------------------------------------------------------------------------------------//

void vAdd(CMat3x3* pm3M, const CMat3x3& m3M1)                         // M+=M1
  {
  (*pm3M)[0][0] += m3M1[0][0];
  (*pm3M)[0][1] += m3M1[0][1];
  (*pm3M)[0][2] += m3M1[0][2];

  (*pm3M)[1][0] += m3M1[1][0];
  (*pm3M)[1][1] += m3M1[1][1];
  (*pm3M)[1][2] += m3M1[1][2];

  (*pm3M)[2][0] += m3M1[2][0];
  (*pm3M)[2][1] += m3M1[2][1];
  (*pm3M)[2][2] += m3M1[2][2];
  }

void vSub(CMat3x3* pm3M, const CMat3x3& m3M1)                         // M-=M1
  {
  (*pm3M)[0][0] -= m3M1[0][0];
  (*pm3M)[0][1] -= m3M1[0][1];
  (*pm3M)[0][2] -= m3M1[0][2];

  (*pm3M)[1][0] -= m3M1[1][0];
  (*pm3M)[1][1] -= m3M1[1][1];
  (*pm3M)[1][2] -= m3M1[1][2];

  (*pm3M)[2][0] -= m3M1[2][0];
  (*pm3M)[2][1] -= m3M1[2][1];
  (*pm3M)[2][2] -= m3M1[2][2];
  }

void vMul(CMat3x3* pm3M, const float fK)                              // M*=k
  {
  (*pm3M)[0][0] *= fK;
  (*pm3M)[0][1] *= fK;
  (*pm3M)[0][2] *= fK;

  (*pm3M)[1][0] *= fK;
  (*pm3M)[1][1] *= fK;
  (*pm3M)[1][2] *= fK;

  (*pm3M)[2][0] *= fK;
  (*pm3M)[2][1] *= fK;
  (*pm3M)[2][2] *= fK;
  }

void vMul(CMat3x3* pm3M, const CMat3x3& m3M1)                         // M*=M1
  {
  *pm3M = (*pm3M)*m3M1;
  }

void vDiv(CMat3x3* pm3M, const float fK)                              // M/=k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;

  (*pm3M)[0][0] *= fKinv;
  (*pm3M)[0][1] *= fKinv;
  (*pm3M)[0][2] *= fKinv;

  (*pm3M)[1][0] *= fKinv;
  (*pm3M)[1][1] *= fKinv;
  (*pm3M)[1][2] *= fKinv;

  (*pm3M)[2][0] *= fKinv;
  (*pm3M)[2][1] *= fKinv;
  (*pm3M)[2][2] *= fKinv;
  }

void vDiv(CMat3x3* pm3M, const CMat3x3& m3M1)                         // M/=M1 (* inverse)
  {
  *pm3M = (*pm3M)*(m3M1.m3Inverse());
  }

void vNeg(CMat3x3* pm3M, const CMat3x3& m3M1)                         // M=-M1
  {
  (*pm3M)[0][0] = -m3M1[0][0];
  (*pm3M)[0][1] = -m3M1[0][1];
  (*pm3M)[0][2] = -m3M1[0][2];

  (*pm3M)[1][0] = -m3M1[1][0];
  (*pm3M)[1][1] = -m3M1[1][1];
  (*pm3M)[1][2] = -m3M1[1][2];

  (*pm3M)[2][0] = -m3M1[2][0];
  (*pm3M)[2][1] = -m3M1[2][1];
  (*pm3M)[2][2] = -m3M1[2][2];
  }

void vMul(CMat3x3* pm3M, const CMat3x3& m3M1,const float fK)          // M=M1*k
  {
  (*pm3M)[0][0] = m3M1[0][0]*fK;
  (*pm3M)[0][1] = m3M1[0][1]*fK;
  (*pm3M)[0][2] = m3M1[0][2]*fK;

  (*pm3M)[1][0] = m3M1[1][0]*fK;
  (*pm3M)[1][1] = m3M1[1][1]*fK;
  (*pm3M)[1][2] = m3M1[1][2]*fK;

  (*pm3M)[2][0] = m3M1[2][0]*fK;
  (*pm3M)[2][1] = m3M1[2][1]*fK;
  (*pm3M)[2][2] = m3M1[2][2]*fK;
  }

void vDiv(CMat3x3* pm3M, const CMat3x3& m3M1,const float fK)          // M=M1/k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;

  (*pm3M)[0][0] = m3M1[0][0]*fKinv;
  (*pm3M)[0][1] = m3M1[0][1]*fKinv;
  (*pm3M)[0][2] = m3M1[0][2]*fKinv;

  (*pm3M)[1][0] = m3M1[1][0]*fKinv;
  (*pm3M)[1][1] = m3M1[1][1]*fKinv;
  (*pm3M)[1][2] = m3M1[1][2]*fKinv;

  (*pm3M)[2][0] = m3M1[2][0]*fKinv;
  (*pm3M)[2][1] = m3M1[2][1]*fKinv;
  (*pm3M)[2][2] = m3M1[2][2]*fKinv;
  }

void vSub(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)     // M=M1-M2
  {
  (*pm3M)[0][0] = m3M1[0][0] - m3M2[0][0];
  (*pm3M)[0][1] = m3M1[0][1] - m3M2[0][1];
  (*pm3M)[0][2] = m3M1[0][2] - m3M2[0][2];

  (*pm3M)[1][0] = m3M1[1][0] - m3M2[1][0];
  (*pm3M)[1][1] = m3M1[1][1] - m3M2[1][1];
  (*pm3M)[1][2] = m3M1[1][2] - m3M2[1][2];

  (*pm3M)[2][0] = m3M1[2][0] - m3M2[2][0];
  (*pm3M)[2][1] = m3M1[2][1] - m3M2[2][1];
  (*pm3M)[2][2] = m3M1[2][2] - m3M2[2][2];
  }

void vAdd(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)     // M=M1+M2
  {
  (*pm3M)[0][0] = m3M1[0][0] + m3M2[0][0];
  (*pm3M)[0][1] = m3M1[0][1] + m3M2[0][1];
  (*pm3M)[0][2] = m3M1[0][2] + m3M2[0][2];

  (*pm3M)[1][0] = m3M1[1][0] + m3M2[1][0];
  (*pm3M)[1][1] = m3M1[1][1] + m3M2[1][1];
  (*pm3M)[1][2] = m3M1[1][2] + m3M2[1][2];

  (*pm3M)[2][0] = m3M1[2][0] + m3M2[2][0];
  (*pm3M)[2][1] = m3M1[2][1] + m3M2[2][1];
  (*pm3M)[2][2] = m3M1[2][2] + m3M2[2][2];
  }

void vMul(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)     // M=M1*M2
  {
  (*pm3M)[0][0] = m3M1[0][0]*m3M2[0][0] + m3M1[0][1]*m3M2[1][0]
                + m3M1[0][2]*m3M2[2][0];
  (*pm3M)[0][1] = m3M1[0][0]*m3M2[0][1] + m3M1[0][1]*m3M2[1][1]
                + m3M1[0][2]*m3M2[2][1];
  (*pm3M)[0][2] = m3M1[0][0]*m3M2[0][2] + m3M1[0][1]*m3M2[1][2]
                + m3M1[0][2]*m3M2[2][2];

  (*pm3M)[1][0] = m3M1[1][0]*m3M2[0][0] + m3M1[1][1]*m3M2[1][0]
                + m3M1[1][2]*m3M2[2][0];
  (*pm3M)[1][1] = m3M1[1][0]*m3M2[0][1] + m3M1[1][1]*m3M2[1][1]
                + m3M1[1][2]*m3M2[2][1];
  (*pm3M)[1][2] = m3M1[1][0]*m3M2[0][2] + m3M1[1][1]*m3M2[1][2]
                + m3M1[1][2]*m3M2[2][2];

  (*pm3M)[2][0] = m3M1[2][0]*m3M2[0][0] + m3M1[2][1]*m3M2[1][0]
                + m3M1[2][2]*m3M2[2][0];
  (*pm3M)[2][1] = m3M1[2][0]*m3M2[0][1] + m3M1[2][1]*m3M2[1][1]
                + m3M1[2][2]*m3M2[2][1];
  (*pm3M)[2][2] = m3M1[2][0]*m3M2[0][2] + m3M1[2][1]*m3M2[1][2]
                + m3M1[2][2]*m3M2[2][2];
  }

void vMulTerm(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)
  {                                                                   // M=M1*M2 (term by term)
  (*pm3M)[0][0] = m3M1[0][0] * m3M2[0][0];
  (*pm3M)[0][1] = m3M1[0][1] * m3M2[0][1];
  (*pm3M)[0][2] = m3M1[0][2] * m3M2[0][2];

  (*pm3M)[1][0] = m3M1[1][0] * m3M2[1][0];
  (*pm3M)[1][1] = m3M1[1][1] * m3M2[1][1];
  (*pm3M)[1][2] = m3M1[1][2] * m3M2[1][2];

  (*pm3M)[2][0] = m3M1[2][0] * m3M2[2][0];
  (*pm3M)[2][1] = m3M1[2][1] * m3M2[2][1];
  (*pm3M)[2][2] = m3M1[2][2] * m3M2[2][2];
  }

void vDiv(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)     // M=M1/M2 (term by term)
  {
  _DDIV0_(m3M2[0][0]);
  _DDIV0_(m3M2[0][1]);
  _DDIV0_(m3M2[0][2]);
  _DDIV0_(m3M2[1][0]);
  _DDIV0_(m3M2[1][1]);
  _DDIV0_(m3M2[1][2]);
  _DDIV0_(m3M2[2][0]);
  _DDIV0_(m3M2[2][1]);
  _DDIV0_(m3M2[2][2]);

  (*pm3M)[0][0] = m3M1[0][0] / m3M2[0][0];
  (*pm3M)[0][1] = m3M1[0][1] / m3M2[0][1];
  (*pm3M)[0][2] = m3M1[0][2] / m3M2[0][2];

  (*pm3M)[1][0] = m3M1[1][0] / m3M2[1][0];
  (*pm3M)[1][1] = m3M1[1][1] / m3M2[1][1];
  (*pm3M)[1][2] = m3M1[1][2] / m3M2[1][2];

  (*pm3M)[2][0] = m3M1[2][0] / m3M2[2][0];
  (*pm3M)[2][1] = m3M1[2][1] / m3M2[2][1];
  (*pm3M)[2][2] = m3M1[2][2] / m3M2[2][2];
  }

void vTranspose(CMat3x3* pm3M, const CMat3x3& m3M1)                   // M=tM1
  {
  (*pm3M)[0][0] = m3M1[0][0];
  (*pm3M)[0][1] = m3M1[1][0];
  (*pm3M)[0][2] = m3M1[2][0];

  (*pm3M)[1][0] = m3M1[0][1];
  (*pm3M)[1][1] = m3M1[1][1];
  (*pm3M)[1][2] = m3M1[2][1];

  (*pm3M)[2][0] = m3M1[0][2];
  (*pm3M)[2][1] = m3M1[1][2];
  (*pm3M)[2][2] = m3M1[2][2];
  }

void vMulMxtM(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)
  {                                                                   // M=M1*tM2
  (*pm3M)[0][0] = m3M1[0][0]*m3M2[0][0] + m3M1[0][1]*m3M2[0][1]
                + m3M1[0][2]*m3M2[0][2];
  (*pm3M)[0][1] = m3M1[0][0]*m3M2[1][0] + m3M1[0][1]*m3M2[1][1]
                + m3M1[0][2]*m3M2[1][2];
  (*pm3M)[0][2] = m3M1[0][0]*m3M2[2][0] + m3M1[0][1]*m3M2[2][1]
                + m3M1[0][2]*m3M2[2][2];

  (*pm3M)[1][0] = m3M1[1][0]*m3M2[0][0] + m3M1[1][1]*m3M2[0][1]
                + m3M1[1][2]*m3M2[0][2];
  (*pm3M)[1][1] = m3M1[1][0]*m3M2[1][0] + m3M1[1][1]*m3M2[1][1]
                + m3M1[1][2]*m3M2[1][2];
  (*pm3M)[1][2] = m3M1[1][0]*m3M2[2][0] + m3M1[1][1]*m3M2[2][1]
                + m3M1[1][2]*m3M2[2][2];

  (*pm3M)[2][0] = m3M1[2][0]*m3M2[0][0] + m3M1[2][1]*m3M2[0][1]
                + m3M1[2][2]*m3M2[0][2];
  (*pm3M)[2][1] = m3M1[2][0]*m3M2[1][0] + m3M1[2][1]*m3M2[1][1]
                + m3M1[2][2]*m3M2[1][2];
  (*pm3M)[2][2] = m3M1[2][0]*m3M2[2][0] + m3M1[2][1]*m3M2[2][1]
                + m3M1[2][2]*m3M2[2][2];
  }

void vMultMxM(CMat3x3* pm3M, const CMat3x3& m3M1,const CMat3x3& m3M2)
  {                                                                   // M=tM1*M2
  (*pm3M)[0][0] = m3M1[0][0]*m3M2[0][0] + m3M1[1][0]*m3M2[1][0]
                + m3M1[2][0]*m3M2[2][0];
  (*pm3M)[0][1] = m3M1[0][0]*m3M2[0][1] + m3M1[1][0]*m3M2[1][1]
                + m3M1[2][0]*m3M2[2][1];
  (*pm3M)[0][2] = m3M1[0][0]*m3M2[0][2] + m3M1[1][0]*m3M2[1][2]
                + m3M1[2][0]*m3M2[2][2];

  (*pm3M)[1][0] = m3M1[0][1]*m3M2[0][0] + m3M1[1][1]*m3M2[1][0]
                + m3M1[2][1]*m3M2[2][0];
  (*pm3M)[1][1] = m3M1[0][1]*m3M2[0][1] + m3M1[1][1]*m3M2[1][1]
                + m3M1[2][1]*m3M2[2][1];
  (*pm3M)[1][2] = m3M1[0][1]*m3M2[0][2] + m3M1[1][1]*m3M2[1][2]
                + m3M1[2][1]*m3M2[2][2];

  (*pm3M)[2][0] = m3M1[0][2]*m3M2[0][0] + m3M1[1][2]*m3M2[1][0]
                + m3M1[2][2]*m3M2[2][0];
  (*pm3M)[2][1] = m3M1[0][2]*m3M2[0][1] + m3M1[1][2]*m3M2[1][1]
                + m3M1[2][2]*m3M2[2][1];
  (*pm3M)[2][2] = m3M1[0][2]*m3M2[0][2] + m3M1[1][2]*m3M2[1][2]
                + m3M1[2][2]*m3M2[2][2];
  }

//--------------------------------------------------------------------------------------------------------------------//
//                                                4x4                                                                 //
//--------------------------------------------------------------------------------------------------------------------//

void vAdd(CMat4x4* pm4M, const CMat4x4& m4M1)                         // M+=M1
  {
  (*pm4M)[0][0] += m4M1[0][0];
  (*pm4M)[0][1] += m4M1[0][1];
  (*pm4M)[0][2] += m4M1[0][2];
  (*pm4M)[0][3] += m4M1[0][3];

  (*pm4M)[1][0] += m4M1[1][0];
  (*pm4M)[1][1] += m4M1[1][1];
  (*pm4M)[1][2] += m4M1[1][2];
  (*pm4M)[1][3] += m4M1[1][3];

  (*pm4M)[2][0] += m4M1[2][0];
  (*pm4M)[2][1] += m4M1[2][1];
  (*pm4M)[2][2] += m4M1[2][2];
  (*pm4M)[2][3] += m4M1[2][3];

  (*pm4M)[3][0] += m4M1[3][0];
  (*pm4M)[3][1] += m4M1[3][1];
  (*pm4M)[3][2] += m4M1[3][2];
  (*pm4M)[3][3] += m4M1[3][3];
  }

void vSub(CMat4x4* pm4M, const CMat4x4& m4M1)                         // M-=M1
  {
  (*pm4M)[0][0] -= m4M1[0][0];
  (*pm4M)[0][1] -= m4M1[0][1];
  (*pm4M)[0][2] -= m4M1[0][2];
  (*pm4M)[0][3] -= m4M1[0][3];

  (*pm4M)[1][0] -= m4M1[1][0];
  (*pm4M)[1][1] -= m4M1[1][1];
  (*pm4M)[1][2] -= m4M1[1][2];
  (*pm4M)[1][3] -= m4M1[1][3];

  (*pm4M)[2][0] -= m4M1[2][0];
  (*pm4M)[2][1] -= m4M1[2][1];
  (*pm4M)[2][2] -= m4M1[2][2];
  (*pm4M)[2][3] -= m4M1[2][3];

  (*pm4M)[3][0] -= m4M1[3][0];
  (*pm4M)[3][1] -= m4M1[3][1];
  (*pm4M)[3][2] -= m4M1[3][2];
  (*pm4M)[3][3] -= m4M1[3][3];
  }

void vMul(CMat4x4* pm4M, const float fK)                              // M*=k
  {
  (*pm4M)[0][0] *= fK;
  (*pm4M)[0][1] *= fK;
  (*pm4M)[0][2] *= fK;
  (*pm4M)[0][3] *= fK;

  (*pm4M)[1][0] *= fK;
  (*pm4M)[1][1] *= fK;
  (*pm4M)[1][2] *= fK;
  (*pm4M)[1][3] *= fK;

  (*pm4M)[2][0] *= fK;
  (*pm4M)[2][1] *= fK;
  (*pm4M)[2][2] *= fK;
  (*pm4M)[2][3] *= fK;

  (*pm4M)[3][0] *= fK;
  (*pm4M)[3][1] *= fK;
  (*pm4M)[3][2] *= fK;
  (*pm4M)[3][3] *= fK;
  }

void vMul(CMat4x4* pm4M, const CMat4x4& m4M1)                         // M*=M1
  {
  *pm4M = (*pm4M)*m4M1;
  }

void vDiv(CMat4x4* pm4M, const float fK)                              // M/=k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;

  (*pm4M)[0][0] *= fKinv;
  (*pm4M)[0][1] *= fKinv;
  (*pm4M)[0][2] *= fKinv;
  (*pm4M)[0][3] *= fKinv;

  (*pm4M)[1][0] *= fKinv;
  (*pm4M)[1][1] *= fKinv;
  (*pm4M)[1][2] *= fKinv;
  (*pm4M)[1][3] *= fKinv;

  (*pm4M)[2][0] *= fKinv;
  (*pm4M)[2][1] *= fKinv;
  (*pm4M)[2][2] *= fKinv;
  (*pm4M)[2][3] *= fKinv;

  (*pm4M)[3][0] *= fKinv;
  (*pm4M)[3][1] *= fKinv;
  (*pm4M)[3][2] *= fKinv;
  (*pm4M)[3][3] *= fKinv;
  }

void vDiv(CMat4x4* pm4M, const CMat4x4& m4M1)                         // M/=M1 (* inverse)
  {
  *pm4M = (*pm4M)*(m4M1.m4Inverse());
  }

void vNeg(CMat4x4* pm4M, const CMat4x4& m4M1)                         // M=-M1
  {
  (*pm4M)[0][0] = -m4M1[0][0];
  (*pm4M)[0][1] = -m4M1[0][1];
  (*pm4M)[0][2] = -m4M1[0][2];
  (*pm4M)[0][3] = -m4M1[0][3];

  (*pm4M)[1][0] = -m4M1[1][0];
  (*pm4M)[1][1] = -m4M1[1][1];
  (*pm4M)[1][2] = -m4M1[1][2];
  (*pm4M)[1][3] = -m4M1[1][3];

  (*pm4M)[2][0] = -m4M1[2][0];
  (*pm4M)[2][1] = -m4M1[2][1];
  (*pm4M)[2][2] = -m4M1[2][2];
  (*pm4M)[2][3] = -m4M1[2][3];

  (*pm4M)[3][0] = -m4M1[3][0];
  (*pm4M)[3][1] = -m4M1[3][1];
  (*pm4M)[3][2] = -m4M1[3][2];
  (*pm4M)[3][3] = -m4M1[3][3];
  }

void vMul(CMat4x4* pm4M, const CMat4x4& m4M1,const float fK)          // M=M1*k
  {
  (*pm4M)[0][0] = m4M1[0][0]*fK;
  (*pm4M)[0][1] = m4M1[0][1]*fK;
  (*pm4M)[0][2] = m4M1[0][2]*fK;
  (*pm4M)[0][3] = m4M1[0][3]*fK;

  (*pm4M)[1][0] = m4M1[1][0]*fK;
  (*pm4M)[1][1] = m4M1[1][1]*fK;
  (*pm4M)[1][2] = m4M1[1][2]*fK;
  (*pm4M)[1][3] = m4M1[1][3]*fK;

  (*pm4M)[2][0] = m4M1[2][0]*fK;
  (*pm4M)[2][1] = m4M1[2][1]*fK;
  (*pm4M)[2][2] = m4M1[2][2]*fK;
  (*pm4M)[2][3] = m4M1[2][3]*fK;

  (*pm4M)[3][0] = m4M1[3][0]*fK;
  (*pm4M)[3][1] = m4M1[3][1]*fK;
  (*pm4M)[3][2] = m4M1[3][2]*fK;
  (*pm4M)[3][3] = m4M1[3][3]*fK;
  }

void vDiv(CMat4x4* pm4M, const CMat4x4& m4M1,const float fK)          // M=M1/k
  {
  _DDIV0_(fK);
  float fKinv = 1.f/fK;

  (*pm4M)[0][0] = m4M1[0][0]*fKinv;
  (*pm4M)[0][1] = m4M1[0][1]*fKinv;
  (*pm4M)[0][2] = m4M1[0][2]*fKinv;
  (*pm4M)[0][3] = m4M1[0][3]*fKinv;

  (*pm4M)[1][0] = m4M1[1][0]*fKinv;
  (*pm4M)[1][1] = m4M1[1][1]*fKinv;
  (*pm4M)[1][2] = m4M1[1][2]*fKinv;
  (*pm4M)[1][3] = m4M1[1][3]*fKinv;

  (*pm4M)[2][0] = m4M1[2][0]*fKinv;
  (*pm4M)[2][1] = m4M1[2][1]*fKinv;
  (*pm4M)[2][2] = m4M1[2][2]*fKinv;
  (*pm4M)[2][3] = m4M1[2][3]*fKinv;

  (*pm4M)[3][0] = m4M1[3][0]*fKinv;
  (*pm4M)[3][1] = m4M1[3][1]*fKinv;
  (*pm4M)[3][2] = m4M1[3][2]*fKinv;
  (*pm4M)[3][3] = m4M1[3][3]*fKinv;
  }

void vSub(CMat4x4* pm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)     // M=M1-M2
  {
  (*pm4M)[0][0] = m4M1[0][0] - m4M2[0][0];
  (*pm4M)[0][1] = m4M1[0][1] - m4M2[0][1];
  (*pm4M)[0][2] = m4M1[0][2] - m4M2[0][2];
  (*pm4M)[0][3] = m4M1[0][3] - m4M2[0][3];

  (*pm4M)[1][0] = m4M1[1][0] - m4M2[1][0];
  (*pm4M)[1][1] = m4M1[1][1] - m4M2[1][1];
  (*pm4M)[1][2] = m4M1[1][2] - m4M2[1][2];
  (*pm4M)[1][3] = m4M1[1][3] - m4M2[1][3];

  (*pm4M)[2][0] = m4M1[2][0] - m4M2[2][0];
  (*pm4M)[2][1] = m4M1[2][1] - m4M2[2][1];
  (*pm4M)[2][2] = m4M1[2][2] - m4M2[2][2];
  (*pm4M)[2][3] = m4M1[2][3] - m4M2[2][3];

  (*pm4M)[3][0] = m4M1[3][0] - m4M2[3][0];
  (*pm4M)[3][1] = m4M1[3][1] - m4M2[3][1];
  (*pm4M)[3][2] = m4M1[3][2] - m4M2[3][2];
  (*pm4M)[3][3] = m4M1[3][3] - m4M2[3][3];
  }

void vAdd(CMat4x4* pm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)     // M=M1+M2
  {
  (*pm4M)[0][0] = m4M1[0][0] + m4M2[0][0];
  (*pm4M)[0][1] = m4M1[0][1] + m4M2[0][1];
  (*pm4M)[0][2] = m4M1[0][2] + m4M2[0][2];
  (*pm4M)[0][3] = m4M1[0][3] + m4M2[0][3];

  (*pm4M)[1][0] = m4M1[1][0] + m4M2[1][0];
  (*pm4M)[1][1] = m4M1[1][1] + m4M2[1][1];
  (*pm4M)[1][2] = m4M1[1][2] + m4M2[1][2];
  (*pm4M)[1][3] = m4M1[1][3] + m4M2[1][3];

  (*pm4M)[2][0] = m4M1[2][0] + m4M2[2][0];
  (*pm4M)[2][1] = m4M1[2][1] + m4M2[2][1];
  (*pm4M)[2][2] = m4M1[2][2] + m4M2[2][2];
  (*pm4M)[2][3] = m4M1[2][3] + m4M2[2][3];

  (*pm4M)[3][0] = m4M1[3][0] + m4M2[3][0];
  (*pm4M)[3][1] = m4M1[3][1] + m4M2[3][1];
  (*pm4M)[3][2] = m4M1[3][2] + m4M2[3][2];
  (*pm4M)[3][3] = m4M1[3][3] + m4M2[3][3];
  }

void vMul(CMat4x4* pm4M,  const CMat4x4& m4M1, const CMat4x4& m4M2)   // M=M1*M2
  {
  (*pm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[0][1]*m4M2[1][0]
                + m4M1[0][2]*m4M2[2][0] + m4M1[0][3]*m4M2[3][0];
  (*pm4M)[0][1] = m4M1[0][0]*m4M2[0][1] + m4M1[0][1]*m4M2[1][1]
                + m4M1[0][2]*m4M2[2][1] + m4M1[0][3]*m4M2[3][1];
  (*pm4M)[0][2] = m4M1[0][0]*m4M2[0][2] + m4M1[0][1]*m4M2[1][2]
                + m4M1[0][2]*m4M2[2][2] + m4M1[0][3]*m4M2[3][2];
  (*pm4M)[0][3] = m4M1[0][0]*m4M2[0][3] + m4M1[0][1]*m4M2[1][3]
                + m4M1[0][2]*m4M2[2][3] + m4M1[0][3]*m4M2[3][3];

  (*pm4M)[1][0] = m4M1[1][0]*m4M2[0][0] + m4M1[1][1]*m4M2[1][0]
                + m4M1[1][2]*m4M2[2][0] + m4M1[1][3]*m4M2[3][0];
  (*pm4M)[1][1] = m4M1[1][0]*m4M2[0][1] + m4M1[1][1]*m4M2[1][1]
                + m4M1[1][2]*m4M2[2][1] + m4M1[1][3]*m4M2[3][1];
  (*pm4M)[1][2] = m4M1[1][0]*m4M2[0][2] + m4M1[1][1]*m4M2[1][2]
                + m4M1[1][2]*m4M2[2][2] + m4M1[1][3]*m4M2[3][2];
  (*pm4M)[1][3] = m4M1[1][0]*m4M2[0][3] + m4M1[1][1]*m4M2[1][3]
                + m4M1[1][2]*m4M2[2][3] + m4M1[1][3]*m4M2[3][3];

  (*pm4M)[2][0] = m4M1[2][0]*m4M2[0][0] + m4M1[2][1]*m4M2[1][0]
                + m4M1[2][2]*m4M2[2][0] + m4M1[2][3]*m4M2[3][0];
  (*pm4M)[2][1] = m4M1[2][0]*m4M2[0][1] + m4M1[2][1]*m4M2[1][1]
                + m4M1[2][2]*m4M2[2][1] + m4M1[2][3]*m4M2[3][1];
  (*pm4M)[2][2] = m4M1[2][0]*m4M2[0][2] + m4M1[2][1]*m4M2[1][2]
                + m4M1[2][2]*m4M2[2][2] + m4M1[2][3]*m4M2[3][2];
  (*pm4M)[2][3] = m4M1[2][0]*m4M2[0][3] + m4M1[2][1]*m4M2[1][3]
                + m4M1[2][2]*m4M2[2][3] + m4M1[2][3]*m4M2[3][3];

  (*pm4M)[3][0] = m4M1[3][0]*m4M2[0][0] + m4M1[3][1]*m4M2[1][0]
                + m4M1[3][2]*m4M2[2][0] + m4M1[3][3]*m4M2[3][0];
  (*pm4M)[3][1] = m4M1[3][0]*m4M2[0][1] + m4M1[3][1]*m4M2[1][1]
                + m4M1[3][2]*m4M2[2][1] + m4M1[3][3]*m4M2[3][1];
  (*pm4M)[3][2] = m4M1[3][0]*m4M2[0][2] + m4M1[3][1]*m4M2[1][2]
                + m4M1[3][2]*m4M2[2][2] + m4M1[3][3]*m4M2[3][2];
  (*pm4M)[3][3] = m4M1[3][0]*m4M2[0][3] + m4M1[3][1]*m4M2[1][3]
                + m4M1[3][2]*m4M2[2][3] + m4M1[3][3]*m4M2[3][3];
  }

void vMulTerm(CMat4x4* pm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)
  {                                                                   // M=M1*M2 (term by term)
  (*pm4M)[0][0] = m4M1[0][0] * m4M2[0][0];
  (*pm4M)[0][1] = m4M1[0][1] * m4M2[0][1];
  (*pm4M)[0][2] = m4M1[0][2] * m4M2[0][2];
  (*pm4M)[0][3] = m4M1[0][3] * m4M2[0][3];

  (*pm4M)[1][0] = m4M1[1][0] * m4M2[1][0];
  (*pm4M)[1][1] = m4M1[1][1] * m4M2[1][1];
  (*pm4M)[1][2] = m4M1[1][2] * m4M2[1][2];
  (*pm4M)[1][3] = m4M1[1][3] * m4M2[1][3];

  (*pm4M)[2][0] = m4M1[2][0] * m4M2[2][0];
  (*pm4M)[2][1] = m4M1[2][1] * m4M2[2][1];
  (*pm4M)[2][2] = m4M1[2][2] * m4M2[2][2];
  (*pm4M)[2][3] = m4M1[2][3] * m4M2[2][3];

  (*pm4M)[3][0] = m4M1[3][0] * m4M2[3][0];
  (*pm4M)[3][1] = m4M1[3][1] * m4M2[3][1];
  (*pm4M)[3][2] = m4M1[3][2] * m4M2[3][2];
  (*pm4M)[3][3] = m4M1[3][3] * m4M2[3][3];
  }

void vDiv(CMat4x4* pm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)    // M=M1/M2 (term by term)
  {
  _DDIV0_(m4M2[0][0]);
  _DDIV0_(m4M2[0][1]);
  _DDIV0_(m4M2[0][2]);
  _DDIV0_(m4M2[0][3]);
  _DDIV0_(m4M2[1][0]);
  _DDIV0_(m4M2[1][1]);
  _DDIV0_(m4M2[1][2]);
  _DDIV0_(m4M2[1][3]);
  _DDIV0_(m4M2[2][0]);
  _DDIV0_(m4M2[2][1]);
  _DDIV0_(m4M2[2][2]);
  _DDIV0_(m4M2[2][3]);
  _DDIV0_(m4M2[3][0]);
  _DDIV0_(m4M2[3][1]);
  _DDIV0_(m4M2[3][2]);
  _DDIV0_(m4M2[3][3]);

  (*pm4M)[0][0] = m4M1[0][0] / m4M2[0][0];
  (*pm4M)[0][1] = m4M1[0][1] / m4M2[0][1];
  (*pm4M)[0][2] = m4M1[0][2] / m4M2[0][2];
  (*pm4M)[0][3] = m4M1[0][3] / m4M2[0][3];

  (*pm4M)[1][0] = m4M1[1][0] / m4M2[1][0];
  (*pm4M)[1][1] = m4M1[1][1] / m4M2[1][1];
  (*pm4M)[1][2] = m4M1[1][2] / m4M2[1][2];
  (*pm4M)[1][3] = m4M1[1][3] / m4M2[1][3];

  (*pm4M)[2][0] = m4M1[2][0] / m4M2[2][0];
  (*pm4M)[2][1] = m4M1[2][1] / m4M2[2][1];
  (*pm4M)[2][2] = m4M1[2][2] / m4M2[2][2];
  (*pm4M)[2][3] = m4M1[2][3] / m4M2[2][3];

  (*pm4M)[3][0] = m4M1[3][0] / m4M2[3][0];
  (*pm4M)[3][1] = m4M1[3][1] / m4M2[3][1];
  (*pm4M)[3][2] = m4M1[3][2] / m4M2[3][2];
  (*pm4M)[3][3] = m4M1[3][3] / m4M2[3][3];
  }

void vTranspose(CMat4x4* pm4M, const CMat4x4& m4M1)                  // M=tM1
  {
  (*pm4M)[0][0] = m4M1[0][0];
  (*pm4M)[0][1] = m4M1[1][0];
  (*pm4M)[0][2] = m4M1[2][0];
  (*pm4M)[0][3] = m4M1[3][0];

  (*pm4M)[1][0] = m4M1[0][1];
  (*pm4M)[1][1] = m4M1[1][1];
  (*pm4M)[1][2] = m4M1[2][1];
  (*pm4M)[1][3] = m4M1[3][1];

  (*pm4M)[2][0] = m4M1[0][2];
  (*pm4M)[2][1] = m4M1[1][2];
  (*pm4M)[2][2] = m4M1[2][2];
  (*pm4M)[2][3] = m4M1[3][2];

  (*pm4M)[3][0] = m4M1[0][3];
  (*pm4M)[3][1] = m4M1[1][3];
  (*pm4M)[3][2] = m4M1[2][3];
  (*pm4M)[3][3] = m4M1[3][3];
  }

void vMulMxtM(CMat4x4* pm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)
  {                                                                   // M=M1*tM2
  (*pm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[0][1]*m4M2[0][1]
                + m4M1[0][2]*m4M2[0][2] + m4M1[0][3]*m4M2[0][3];
  (*pm4M)[0][1] = m4M1[0][0]*m4M2[1][0] + m4M1[0][1]*m4M2[1][1]
                + m4M1[0][2]*m4M2[1][2] + m4M1[0][3]*m4M2[1][3];
  (*pm4M)[0][2] = m4M1[0][0]*m4M2[2][0] + m4M1[0][1]*m4M2[2][1]
                + m4M1[0][2]*m4M2[2][2] + m4M1[0][3]*m4M2[2][3];
  (*pm4M)[0][3] = m4M1[0][0]*m4M2[3][0] + m4M1[0][1]*m4M2[3][1]
                + m4M1[0][2]*m4M2[3][2] + m4M1[0][3]*m4M2[3][3];

  (*pm4M)[1][0] = m4M1[1][0]*m4M2[0][0] + m4M1[1][1]*m4M2[0][1]
                + m4M1[1][2]*m4M2[0][2] + m4M1[1][3]*m4M2[0][3];
  (*pm4M)[1][1] = m4M1[1][0]*m4M2[1][0] + m4M1[1][1]*m4M2[1][1]
                + m4M1[1][2]*m4M2[1][2] + m4M1[1][3]*m4M2[1][3];
  (*pm4M)[1][2] = m4M1[1][0]*m4M2[2][0] + m4M1[1][1]*m4M2[2][1]
                + m4M1[1][2]*m4M2[2][2] + m4M1[1][3]*m4M2[2][3];
  (*pm4M)[1][3] = m4M1[1][0]*m4M2[3][0] + m4M1[1][1]*m4M2[3][1]
                + m4M1[1][2]*m4M2[3][2] + m4M1[1][3]*m4M2[3][3];

  (*pm4M)[2][0] = m4M1[2][0]*m4M2[0][0] + m4M1[2][1]*m4M2[0][1]
                + m4M1[2][2]*m4M2[0][2] + m4M1[2][3]*m4M2[0][3];
  (*pm4M)[2][1] = m4M1[2][0]*m4M2[1][0] + m4M1[2][1]*m4M2[1][1]
                + m4M1[2][2]*m4M2[1][2] + m4M1[2][3]*m4M2[1][3];
  (*pm4M)[2][2] = m4M1[2][0]*m4M2[2][0] + m4M1[2][1]*m4M2[2][1]
                + m4M1[2][2]*m4M2[2][2] + m4M1[2][3]*m4M2[2][3];
  (*pm4M)[2][3] = m4M1[2][0]*m4M2[3][0] + m4M1[2][1]*m4M2[3][1]
                + m4M1[2][2]*m4M2[3][2] + m4M1[2][3]*m4M2[3][3];

  (*pm4M)[3][0] = m4M1[3][0]*m4M2[0][0] + m4M1[3][1]*m4M2[0][1]
                + m4M1[3][2]*m4M2[0][2] + m4M1[3][3]*m4M2[0][3];
  (*pm4M)[3][1] = m4M1[3][0]*m4M2[1][0] + m4M1[3][1]*m4M2[1][1]
                + m4M1[3][2]*m4M2[1][2] + m4M1[3][3]*m4M2[1][3];
  (*pm4M)[3][2] = m4M1[3][0]*m4M2[2][0] + m4M1[3][1]*m4M2[2][1]
                + m4M1[3][2]*m4M2[2][2] + m4M1[3][3]*m4M2[2][3];
  (*pm4M)[3][3] = m4M1[3][0]*m4M2[3][0] + m4M1[3][1]*m4M2[3][1]
                + m4M1[3][2]*m4M2[3][2] + m4M1[3][3]*m4M2[3][3];
  }

void vMultMxM(CMat4x4* pm4M, const CMat4x4& m4M1,const CMat4x4& m4M2)
  {                                                                   // M=tM1*M2
  (*pm4M)[0][0] = m4M1[0][0]*m4M2[0][0] + m4M1[1][0]*m4M2[1][0]
                + m4M1[2][0]*m4M2[2][0] + m4M1[3][0]*m4M2[3][0];
  (*pm4M)[0][1] = m4M1[0][0]*m4M2[0][1] + m4M1[1][0]*m4M2[1][1]
                + m4M1[2][0]*m4M2[2][1] + m4M1[3][0]*m4M2[3][1];
  (*pm4M)[0][2] = m4M1[0][0]*m4M2[0][2] + m4M1[1][0]*m4M2[1][2]
                + m4M1[2][0]*m4M2[2][2] + m4M1[3][0]*m4M2[3][2];
  (*pm4M)[0][3] = m4M1[0][0]*m4M2[0][3] + m4M1[1][0]*m4M2[1][3]
                + m4M1[2][0]*m4M2[2][3] + m4M1[3][0]*m4M2[3][3];

  (*pm4M)[1][0] = m4M1[0][1]*m4M2[0][0] + m4M1[1][1]*m4M2[1][0]
                + m4M1[2][1]*m4M2[2][0] + m4M1[3][1]*m4M2[3][0];
  (*pm4M)[1][1] = m4M1[0][1]*m4M2[0][1] + m4M1[1][1]*m4M2[1][1]
                + m4M1[2][1]*m4M2[2][1] + m4M1[3][1]*m4M2[3][1];
  (*pm4M)[1][2] = m4M1[0][1]*m4M2[0][2] + m4M1[1][1]*m4M2[1][2]
                + m4M1[2][1]*m4M2[2][2] + m4M1[3][1]*m4M2[3][2];
  (*pm4M)[1][3] = m4M1[0][1]*m4M2[0][3] + m4M1[1][1]*m4M2[1][3]
                + m4M1[2][1]*m4M2[2][3] + m4M1[3][1]*m4M2[3][3];

  (*pm4M)[2][0] = m4M1[0][2]*m4M2[0][0] + m4M1[1][2]*m4M2[1][0]
                + m4M1[2][2]*m4M2[2][0] + m4M1[3][2]*m4M2[3][0];
  (*pm4M)[2][1] = m4M1[0][2]*m4M2[0][1] + m4M1[1][2]*m4M2[1][1]
                + m4M1[2][2]*m4M2[2][1] + m4M1[3][2]*m4M2[3][1];
  (*pm4M)[2][2] = m4M1[0][2]*m4M2[0][2] + m4M1[1][2]*m4M2[1][2]
                + m4M1[2][2]*m4M2[2][2] + m4M1[3][2]*m4M2[3][2];
  (*pm4M)[2][3] = m4M1[0][2]*m4M2[0][3] + m4M1[1][2]*m4M2[1][3]
                + m4M1[2][2]*m4M2[2][3] + m4M1[3][2]*m4M2[3][3];

  (*pm4M)[3][0] = m4M1[0][3]*m4M2[0][0] + m4M1[1][3]*m4M2[1][0]
                + m4M1[2][3]*m4M2[2][0] + m4M1[3][3]*m4M2[3][0];
  (*pm4M)[3][1] = m4M1[0][3]*m4M2[0][1] + m4M1[1][3]*m4M2[1][1]
                + m4M1[2][3]*m4M2[2][1] + m4M1[3][3]*m4M2[3][1];
  (*pm4M)[3][2] = m4M1[0][3]*m4M2[0][2] + m4M1[1][3]*m4M2[1][2]
                + m4M1[2][3]*m4M2[2][2] + m4M1[3][3]*m4M2[3][2];
  (*pm4M)[3][3] = m4M1[0][3]*m4M2[0][3] + m4M1[1][3]*m4M2[1][3]
                + m4M1[2][3]*m4M2[2][3] + m4M1[3][3]*m4M2[3][3];
  }
